# Written by Alexander Belchenko
# based on Robert Collins code
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Show all 'heads' in a repository"""


import os
import time

from bzrlib.commands import Command, display_command, register_command
from bzrlib.option import Option


class cmd_heads(Command):
    """Show all revisions in a repository not having descendants.
    """
    takes_options = [Option('by-date', help='Sort heads by date (descending)'),
                     Option('all', help='Show all heads (dead and alive)'),
                     Option('dead-only', help='Show only dead heads'),
                     Option('debug-time', help='Enable debug print of operations times'),
                    ]

    encoding_type = "replace"

    @display_command
    def run(self, by_date=False, all=False, dead_only=False, debug_time=False):
        from bzrlib import errors
        from bzrlib.osutils import format_date
        import bzrlib.repository

        self._init_elapsed_time(debug_time)

        to_file = self.outf

        branch = None
        try:
            branch = bzrlib.branch.Branch.open_containing('.')[0]
            repo = branch.repository
        except errors.NotBranchError:
            try:
                repo = bzrlib.repository.Repository.open('.')
            except errors.NotBranchError:
                print >>to_file, \
                      ("You need to run this command "
                       "either from the root of a shared repository,\n"
                       "or from a branch.")
                return 3

        g = repo.get_revision_graph()
        possible_heads = set(g.keys())
        not_heads = set()
        for parents in g.values():
            not_heads.update(set(parents))
        
        heads = possible_heads.difference(not_heads)

        # TODO: use different sorting schemes instead of alphabetical sort
        heads = list(heads)
        heads.sort()

        self._print_elapsed_time('get heads:')

        # sorting by date
        if by_date:
            dates = {}
            for head in heads:
                rev = repo.get_revision(head)
                timestamp = rev.timestamp
                dates[timestamp] = head
            keys = dates.keys()
            keys.sort()
            keys.reverse()
            heads = []
            for k in keys:
                heads.append(dates[k])
            self._print_elapsed_time('sort by date:')

        ## mark heads as dead or alive
        # mark all heads as dead
        head_mark = {}
        for head in heads:
            head_mark[head] = 'dead'
        # give the list of live branches in repository or current branch tip
        if not branch:
            # runs from shared repo root
            for directory in os.listdir(u'.'):
                if directory == '.bzr':
                    continue
                if not os.path.isdir(os.path.join(directory, '.bzr')):
                    continue
                try:
                    b, path = bzrlib.branch.Branch.open_containing(directory)
                except errors.NotBranchError:
                    continue
                last_revid = b.last_revision()
                head_mark[last_revid] = 'alive'
#                self._print_elapsed_time('proceed branch ' + str(b)+':')
        else:
            # runs from branch
            # TODO: implementation
            raise NotImplementedError('TODO: need to detect is there is shared repo or not')
        self._print_elapsed_time('make head marks:')

        # show time
        indent = ' '*2
        show_timezone = 'original'
        
        for head in heads:

            mark = head_mark[head]
            if not all:
                if dead_only:
                    if mark != 'dead':
                        continue
                else:
                    if mark != 'alive':
                        continue
                    
            print >>to_file, "revision-id:", head,
            if mark == 'dead':  print >>to_file, '(dead)'
            else:               print >>to_file

            rev = repo.get_revision(head)
            # borrowed from LongLogFormatter
            print >>to_file,  indent+'committer:', rev.committer
            try:
                print >>to_file, indent+'branch nick: %s' % \
                    rev.properties['branch-nick']
            except KeyError:
                pass
            date_str = format_date(rev.timestamp,
                                   rev.timezone or 0,
                                   show_timezone)
            print >>to_file,  indent+'timestamp: %s' % date_str
    
            print >>to_file,  indent+'message:'
            if not rev.message:
                print >>to_file,  indent+'  (no message)'
            else:
                message = rev.message.rstrip('\r\n')
                for l in message.split('\n'):
                    print >>to_file,  indent+'  ' + l
            self._print_elapsed_time('print head:')
            print

        if not heads:
            print >>to_file, 'No heads found'
            return 1

    def _init_elapsed_time(self, debug_time=False):
        self.debug_time = debug_time
        if debug_time:
            self._time = time.time()

    def _print_elapsed_time(self, msg):
        if self.debug_time:
            last_time = time.time()
            print msg, last_time - self._time
            self._time = last_time
#/class cmd_heads


register_command(cmd_heads)
